/*
#
#    This distribution makes makes asynchronous client/server communication possible 
#    using TLVs (Type Length Value). It achieves that letting the user define
#    an IDL which serves as input to a code generator.  The code generated is linked in
#    the client and the server program.
#
#    Copyright (C) <2005>  <Sanjay Jaiman>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
*/

/*_############################################################################
  _## 
  _##  SocketInterface.cpp  
  _## 
  _##
  _##  
  _##########################################################################*/


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/unistd.h>
#include <iostream.h>
#include <socketInterface.h>
#include <syslog.h>
#include <errno.h>
#include "globals.h"

extern int errno;

#define    TWENTYK   20480

transaction *transaction::T_head = NULL;
transaction *transaction::T_tail = NULL;
int  transaction::globalTid = 1;
static char traceBuff[512];

#define SOCKETS_DEBUG ((trace_level > level && (subSystem_trace_mask & SOCKET_LIB_TRACE)) ? true : false)

static void waitForInMilliSeconds(int milliSecs)
{
    int status;
    struct timeval t;
    t.tv_sec = milliSecs / 1000;
    t.tv_usec = (milliSecs % 1000)*1000;
    status = select(0,0,0,0,&t);
}

//-----------------------------------------------------------------------------
//
//    SOCKMUTEX
//
//-----------------------------------------------------------------------------

// timeout in secs
int  SockMutex::timedWait(int timeout)
{
  int ret = 0;
  for(int i=0; i<timeout*100; i++) {
    if (sem_trywait(&mutex) == -1) {
      if (errno == EAGAIN) {
	//	cout << "SockMutex::Waiting...." << endl;
	waitForInMilliSeconds(10);
      }
    }
    else {
      //      cout << "SockMutex::Exiting wait loop" << endl;
      ret = 1;
      break;
    }
  }
  //  cout << "SockMutex::timedWait::returning " << ret << endl;
  return ret;
};

//-----------------------------------------------------------------------------
//
//    TRANSACTION
//
//-----------------------------------------------------------------------------

transaction::transaction(int t_out) : next(NULL), prev(NULL)
{
  if (globalTid == 0x0000ffff)
    globalTid = 1;  // Wrap
  tid = globalTid++;
  //  cout << "new transaction " << tid << endl;
  prev = T_tail; 
  if (T_tail) {
    T_tail->next = this; T_tail = this;
  }
  if (!T_head) {
    T_head = T_tail = this;
  }
  timeOut = t_out;
}

int transaction::timedWait()
{
  if (timeOut)
    return first.timedWait(timeOut);
  else
    wait();
  return 1;
};

transaction::~transaction()
{
  //  cout << "deleting transaction " << tid << endl;
  if (prev) {prev->next = next;}
  if (next) {next->prev = prev;}
  if (T_head == this) {T_head = next;}
  if (T_tail == this) {T_tail = prev;}
}
transaction *transaction::findtid(int t) {
  for (transaction *p = T_head; p; p = p->next)
    if (p->tid == t)
      return p;
  return NULL;
};

//-----------------------------------------------------------------------------
//
//    SENDBUFFERLIST
//
//-----------------------------------------------------------------------------


sendBufferList::~sendBufferList()
{
  if (bufferHead) {
    bufferHead->freeRecursive();
    bufferHead = NULL;
  }
};  // Removes the entire list

int sendBufferList::add(char *buff, int buffLen)
{

  if (numBuffers == MAX_BUFFER_LIMIT) {
    sprintf(traceBuff, "sendBufferList::add::Max buffer limit reached\n");
    PRINTTRACE(2, SOCKET_LIB_TRACE, traceBuff);
    return 0;
  }
  sendBuffer *sb = new sendBuffer(buff, buffLen);
  if (!sb) {
    sprintf(traceBuff, "sendBufferList::add::Unable to allocate memory for new buffer\n");
    PRINTTRACE(2, SOCKET_LIB_TRACE, traceBuff);
    return 0;
  }
  sb->prev = bufferTail; 
  if (bufferTail) {bufferTail->next = sb; bufferTail = sb;}
  if (!bufferHead) {bufferHead = bufferTail = sb;}
  numBuffers++;
  sprintf(traceBuff, "sendBufferList::add:: Number of buffers = %d\n", numBuffers);
  PRINTTRACE(3, SOCKET_LIB_TRACE, traceBuff);
  return 1;
};

sendBuffer *sendBufferList::remove(sendBuffer *sb)
{
  // returns the next one

  if (sb->prev) {sb->prev->next = sb->next;}
  if (sb->next) {sb->next->prev = sb->prev;}
  if (bufferHead == sb) {bufferHead = sb->next;}
  if (bufferTail == sb) {bufferTail = sb->prev;}
  numBuffers--;
  sprintf(traceBuff, "sendBufferList::remove:: Number of buffers = %d\n", numBuffers);
  PRINTTRACE(3, SOCKET_LIB_TRACE, traceBuff);
  return sb->next;
}

// This is to get rid of any/all send buffers that may be queued up
void sendBufferList::cleanup()
{
  if (bufferHead)
    bufferHead->freeRecursive();
  bufferHead = bufferTail = NULL;
  numBuffers = 0;
};

// 0 = success
// 1 = EAGAIN  - try again later
// 2 = FAILURE
int sendBufferList::flushToSocket(int fd)
{
  int level = 6;
  bool debug = SOCKETS_DEBUG;

  if (debug) {
    sprintf(traceBuff, "\tflushToSocket:: (%d)\n", fd);
    printDebug(traceBuff, strlen(traceBuff));
  }
  if (fd == -1) {
    PRINTTRACE(6, SOCKET_LIB_TRACE, "\t\treturning error\n");
    return 2;
  }

  for (sendBuffer *s_b = bufferHead; s_b;) {
    int n = write(fd, s_b->buffer, s_b->bufferLen);   // WRITE
    if (n < 0) {
      switch(errno) {
      case EBADF:
      case EINVAL:
      case EFAULT:
      case EFBIG:
      case EPIPE:
	break;
      case EAGAIN:
	if (debug) {
	  sprintf(traceBuff, "\t\tEAGAIN HIT - SOCKET FULL %d\n", fd);
	  printDebug(traceBuff, strlen(traceBuff));
	}
	return 1;
      case EINTR:
      case ENOSPC:
      case EIO:
	break;
      }
      if (debug) {
	sprintf(traceBuff, "\t\twrite failed on %d - returning error\n", fd);
      }
      perror("SocketInterface::flushToSocket::Write Failed");
      return 2;  //Error
    }
    else if (n < s_b->bufferLen) {   // partial write
      s_b->bufferLen = s_b->bufferLen - n;
      memmove(s_b->buffer, s_b->buffer + n, s_b->bufferLen); // move unwritten data up
      s_b->buffer = (char *)realloc(s_b->buffer, s_b->bufferLen);
      if (!s_b->buffer) {
	cout << "sendBufferList::flushToSocket::Unable to allocate memory - exiting" << endl;
	syslog(LOG_CRIT, "socketInterface:Unable to allocate memory - exiting");
	exit(0);
      }
      if (debug) {
	sprintf(traceBuff, "\t\tPartial Write - SOCKET FULL %d\n", fd);
	printDebug(traceBuff, strlen(traceBuff));
      }
      return 1;
      break;
    }
    else { // full write
      sendBuffer *n = s_b;
      s_b = remove(s_b); // next ptr
      delete n;
    }
  } // for
  PRINTTRACE(level, SOCKET_LIB_TRACE, "\t\tSuccess\n");
  return 0;
}

//-----------------------------------------------------------------------------
//
//    SOCKETINTERFACE
//
//-----------------------------------------------------------------------------

int SocketInterface::verifyIpAddress(char *str)
{
  int i1, i2, i3, i4;
  if (sscanf(str, "%d.%d.%d.%d", &i1, &i2, &i3, &i4) == 4) {
    if (i1 >= 0 && i1 <= 255 &&
	i2 >= 0 && i2 <= 255 &&
	i3 >= 0 && i3 <= 255 &&
	i4 >= 0 && i4 <= 255)
      return 1;
  }
  return 0;
}

// returns:
//       0: Success
//       1: Called with NULL connection object
//       2: Already connected
//       3: Connect Pending
//       4: Unable to realove hostname
//       5: Connect in progress or failed
// If 0 is returned, the user is expected to call "userConnectionEstablihed" on the "conn" object
int SocketInterface::connectToServer(int port, Connection *conn, char *ipAddress, char *errorString)
{
  struct sockaddr_in serv_addr;
  if (!conn) {
    sprintf(errorString, "connectToServer called with NULL connection object\n");
    return 1;
  }
  if (conn->isConnected()) {
    sprintf(errorString, "connectToServer - Already connected to %s\n", ipAddress);
    return 2;
  }
  if (conn->isConnectPending()) {
    time_t t;
    time(&t);
    if (t - conn->lastConnectTime < 5) {
      sprintf(errorString, "connectToServer - connect pending to %s\n", ipAddress);
      return 3;
    }
    // 5 seconds elapsed since last connect - try the connect again
  }
  bzero((void *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;

  if (verifyIpAddress(ipAddress)) {
    serv_addr.sin_addr.s_addr = inet_addr(ipAddress);
  }
  else {
    struct hostent *host = gethostbyname(ipAddress);
    if (host == NULL) {
      sprintf(errorString, "Unable to resolve hostname %s to ipaddress\n", ipAddress);
      return 4;
    }
    struct in_addr *s = ( struct in_addr *)host->h_addr_list[0];
    serv_addr.sin_addr.s_addr = s->s_addr;
    strcpy(ipAddress, inet_ntoa(*s));
    sprintf(errorString, "Connecting to %s\n", ipAddress);
  }

  serv_addr.sin_port = htons(port);
  sprintf(errorString, "Trying to connect to %s on port %d\n", ipAddress, port);
  int sockFd = conn->getFd();
  if (sockFd == -1) {
    sprintf(errorString, "Socket found to be -1; recreating a new non-blocking socket\n");
    sockFd = SocketInterface::createNonBlockingSocket();
    conn->setFd(sockFd);
    sprintf(errorString, "New non-blocking socket created; fd = %d\n", sockFd);
  }
  if (!conn->Connect((struct sockaddr *)&serv_addr)) {
    if (conn->isConnectPending()) {
      sprintf(errorString, "connectToServer - connect pending to %s\n", ipAddress);
      return 3;
    }
    sprintf(errorString, "Connect to %s failed - try later\n", ipAddress);
    return 5;
  }
  conn->connected();
  conn->userConnectionEstablished();
  return 0;
}

int SocketInterface::createNonBlockingSocket()
{
  int sockFd;
  if ((sockFd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    return -1;
  }
  // Make the socket non-blocking
  int flags = fcntl(sockFd, F_GETFL, 0);
  fcntl(sockFd, F_SETFL, flags | O_NONBLOCK);
  struct linger soLinger = {1, 0};  // abort the tcp connection
  setsockopt(sockFd, SOL_SOCKET, SO_LINGER, (void *)&soLinger, sizeof(soLinger));
  return sockFd;
}

int SocketInterface::acceptAndGetNewSocket(struct sockaddr &peer, int listenFd)
{
  int    newFd = -1;

  socklen_t       addrlen = sizeof(struct sockaddr);
  newFd = accept(listenFd, &peer, &addrlen);
  if (newFd < 0) {
    cout << "Error in accept call" << endl;
    perror("accept");
    return -1;
  }
  // Make the socket non-blocking
  int flags = fcntl(newFd, F_GETFL, 0);
  fcntl(newFd, F_SETFL, flags | O_NONBLOCK);

  struct linger soLinger = {1, 0};  // abort the tcp connection
  setsockopt(newFd, SOL_SOCKET, SO_LINGER, (void *)&soLinger, sizeof(soLinger));

  return newFd;
}

void SocketInterface::socketCleanup()
{
  if (readBuffer) free (readBuffer);
  readBuffer = processedDataPtr = NULL;
  processedDataLen = readBufferLen = 0;
  receiveTlv.flush();
  m_sendBuffers.cleanup();
  if (fd != -1) {
    //    cout << "SocketInterface::Closing socket " << fd << endl;
    ::close(fd);
  }
  fd = -1;
  if (isServer()) {
    // cout << "Marking Connection DEAD" << endl;
    State = DEAD; // Mark this one unusable
  }
  else {
    State = DISCONNECTED;  // Still usable for "client" case
  }
};

bool SocketInterface::bufferExceeded()
{
  return (readBufferLen > TWENTYK);  // Throttle at 20K
}

bool SocketInterface::read2K()
{
  bool returnCode = true;
  char *readPtr;
  int  nread;
  char *tPtr = traceBuff;
  int level = 3;
  
  bool debug = SOCKETS_DEBUG;
  if (debug) {
    tPtr += sprintf(tPtr, "read2K:: fd = %d\n", fd);
  }

  // First time through readBuffer is NULL thus this will act as a malloc
  // hence forth readBuffer will keep on growing and shrinking
  // "readBuffer" should never be altered by this routine
  readBuffer = (char *)realloc(readBuffer, readBufferLen + 2048);
  //  printf("SocketInterface::read2K 1::realloc::readbuffer = %x\n", readBuffer);
  if (!readBuffer) {
    goto errorexit;
  }
  //Calculate and re-position the processedDataPtr
  processedDataPtr = readBuffer + processedDataLen;

  readPtr = readBuffer + readBufferLen;  // Used to read new data from the socket 
  //  printf ("SOCKETINTERFACE:: Sizeof Read Buffer:: %d\n", readBufferLen + 2048);
  nread = ::read(fd, readPtr, 2048);
  if (nread < 0) {  //Error
    int locErr = errno;
    if (debug) {
      tPtr += sprintf(tPtr, "\tRead ERROR::Errno = %d Str = %s\n", locErr, sys_errlist[locErr]);
    }
    if (locErr == EINTR || locErr == EAGAIN || locErr == ENOTCONN) {
      if (debug) {
	tPtr += sprintf(tPtr, "\tRead returned EINTR/EAGAIN/ENOTCONN\n");
      }
      // Don't disconnect in this case
    }
    else {
      perror("READ Error");
      if (debug) {
	tPtr += sprintf(tPtr, "\tError reading from socket:: %d - Disconnecting\n", fd);
      }
      disconnected();  // This frees up the readBuffer therefore cannot be combined with the following else if
      returnCode = false;
    }
  }
  else if (nread == 0) {
    // No data read - realloc back to orignal buffer
    //    printf ("SOCKETINTERFACE:: No data read from socket:: %d\n", fd);
    // Used to be like this(without the disconnected call) -
    // readBuffer = (char *)realloc(readBuffer, readBufferLen-2048);
    if (debug) {
      tPtr += sprintf(tPtr, "\tread 0 bytes from socket:: %d - Disconnecting\n", fd);
    }
    disconnected();  
    returnCode = false;
  }
  else if (nread != 2048) {
    // Free up unused space - this can be optimised
    if (debug) {
      tPtr += sprintf(tPtr,"\tread %d bytes off the socket\n", nread);
    }
    readBuffer = (char *)realloc(readBuffer, readBufferLen + nread);
    if (!readBuffer) {
      goto errorexit;
    }
    if (debug) {
      tPtr += sprintf(tPtr,"\trealloc::readbuffer = %x; readBufferLen = %d\n",
		 (unsigned int)readBuffer, readBufferLen);
    }
    readBufferLen += nread;
  }
  else {
    readBufferLen += 2048;
  }
  if (debug) {
    tPtr += sprintf(tPtr, "\treturning %s\n", ((returnCode) ? "true" : "false"));
    printDebug(traceBuff, strlen(traceBuff));
  }
  return returnCode;
  
 errorexit:
  if (debug) {
    tPtr += sprintf(tPtr,"sendBufferList::read2K::Unable to allocate memory - exiting\n");
    printDebug(traceBuff, strlen(traceBuff));
  }
  syslog(LOG_CRIT, "socketInterface:Unable to allocate memory - exiting");
  exit(0);
}


int SocketInterface::shrinkReadBuffer()
{
  int ret = 1;
  char *tPtr = traceBuff;
  int fd = getFd();
  int level = 5;
  bool debug = SOCKETS_DEBUG;

  if (debug) {
    tPtr += sprintf(tPtr, "shrinkReadBuffer:: fd = %d;\n", fd);
  }
  if (receiveTlv.getState() == tlv::TLV_COMPLETE) {
    // We get here if the tlv has been processed which is being held in this object
    // It is absolutely critical we don't call tlv::flush here as the receive tlv could be still 
    // be looked at some other thread (if a blocking call was made).  In which case, the free
    // of memory in flush() would make that thread get a SEGV violation. Instead just mark the TLV
    // as being in the "type" state to start receiving the next tlv
    //    receiveTlv.flush();
    int initLen = readBufferLen;
    if (debug) {
      tPtr += sprintf(tPtr, "\tRemoving a full TLV from the read buffer\n");
    }
    receiveTlv.reinit();
    int dataLeftOver = readBufferLen - processedDataLen;
    if (dataLeftOver == 0) {
      free(readBuffer);
      readBuffer = NULL;
    }
    else {
      memmove(readBuffer, processedDataPtr, dataLeftOver);
      readBuffer = (char *)realloc(readBuffer, dataLeftOver); // Shrink allocated space
      if (!readBuffer) {
	cout << "SocketInterface::shrinkReadBuffer::ERROR in realloc - exiting" << endl;
	syslog(LOG_CRIT, "socketInterface:Unable to allocate memory - exiting");
	exit(0);
      }
    }
    readBufferLen = dataLeftOver;
    if (readBufferLen < 0) {
      if (debug) {
	tPtr += sprintf(tPtr, "\tERROR readBufferLen -ve; returning FAILURE\n");
      }
      ret = 0;
    }
    else {
      if (debug) {
	tPtr += sprintf(tPtr, "\trealloc::shrunk read buffer from %d bytes to %d bytes; new value of read ptr %x\n", 
			initLen, readBufferLen, (unsigned int)readBuffer);
      }
      processedDataLen = 0;
      processedDataPtr = readBuffer;
    }
  }
  else {
    if (debug) {
      tPtr += sprintf(tPtr, "\tTLV incomplete; Nothing to be done\n");
    }
  }
  PRINTTRACE(level, SOCKET_LIB_TRACE, traceBuff);
  return ret;
}

bool SocketInterface::getTlv()
{
  unsigned int bufLen = readBufferLen - processedDataLen;
  int tlvRet = receiveTlv.build(processedDataPtr, bufLen); // processedDataPtr & bufLen come back modified
  processedDataLen = processedDataPtr - readBuffer; 
  if (tlvRet == tlv::TLV_COMPLETE) {
    // Managed to get a full tlv
    return true;
  }
  return false;
}

bool SocketInterface::buffersToSend()
{
  if (m_sendBuffers.numBuffers) {
    return true;
  }
  return false;
};

bool SocketInterface::flush()
{
  bool retCode = true;
  int level = 5;
  bool debug = SOCKETS_DEBUG;

  if (buffersToSend()) {
    if (debug) {
      sprintf(traceBuff, "\tflush()::(%d):: There are buffers to send on the socket; flushing...\n", fd);
      printDebug(traceBuff, strlen(traceBuff));
    }
    int ret = m_sendBuffers.flushToSocket(fd);
    if (ret == 1) {
      State = SOCKET_FULL;
    }
    else if(ret == 2) {
      retCode = false;
    }
  }
  return retCode;
}

bool SocketInterface::Send(char *buff, int buffLen)
{
  int level = 3;
  bool debug = SOCKETS_DEBUG;
  bool returnCode = false;
  int n;

  if (debug) {
    sprintf(traceBuff, "Send:: fd = %d\n", fd);
    printDebug(traceBuff, strlen(traceBuff));
  }
  
  if (fd == -1) 
    goto exit;

  if (!flush()) {
    if (debug) {
      sprintf(traceBuff, "\tError - flush failed - disconnecting\n");
      printDebug(traceBuff, strlen(traceBuff));
    }
    disconnected();
    goto exit;
  }

  if (buffersToSend()) {
    if (m_sendBuffers.add(buff, buffLen))  { // put this buffer in queue
      returnCode = true;
    }
    if (debug) {
      sprintf(traceBuff, "\t%s to send buffer list\n", ((returnCode) ? "Unable to add" : "Added"));
      printDebug(traceBuff, strlen(traceBuff));
    }
    goto exit;
  }
  n = write(fd, buff, buffLen);   // WRITE
  if (n < 0) {
    switch(errno) {
    case EAGAIN:
      if (debug) {
	sprintf(traceBuff, "\tEAGAIN HIT; fd = %d\n", fd);
	printDebug(traceBuff, strlen(traceBuff));
      }
      State = SOCKET_FULL;
      break;
    case EBADF:
    case EINVAL:
    case EFAULT:
    case EFBIG:
    case EPIPE:
    case EINTR:
    case ENOSPC:
    case EIO:
    default:
      if (debug) {
	sprintf(traceBuff, "\tError - Write Failed - disconnecting\n");
	printDebug(traceBuff, strlen(traceBuff));
      }
      perror("SocketInterface::Send::Write Failed");
      disconnected();
      goto exit;
    }
    n = 0;  //To buffer this one ...
  }
  returnCode = true;
  if (n < buffLen) {
    if (!m_sendBuffers.add(buff + n, buffLen - n)) {// put unwritten buffer in queue
      returnCode = false;
    }
    if (debug) {
      sprintf(traceBuff, "\tPartial write\n");
      sprintf(traceBuff, "\t%s to send buffer list\n", ((returnCode) ? "Unable to add" : "Added"));
      printDebug(traceBuff, strlen(traceBuff));
    }
    goto exit;
  }
  // All bytes written

 exit:
  if (debug) {
    sprintf(traceBuff, "\treturning %s\n", ((returnCode) ? "Success" : "Failure"));
    printDebug(traceBuff, strlen(traceBuff));
  }
  return returnCode;
  
}

bool SocketInterface::sendTlv(tlv &TLV)
{
  int type = TLV.getType();
  int length = TLV.getLength();
  char *val = TLV.getVal();
  return sendTlv(type, length, val);
}

bool SocketInterface::sendTlv(int type, int length, char *val)
{
  char TL[2*sizeof(int)];
  char *p = TL;

  // We do two sends here to avoid mallocing huge amounts of data
  // which would be stored anyway if partial send occurred
  int l = 2*sizeof(int);
  bool ret = true;
  tlv::encodeTL(p, type, length);  // This can never fail
  ret = Send(TL, l);
  if (ret) {
    ret = Send(val, length);
    if (!ret) {
      PRINTTRACE(2, SOCKET_LIB_TRACE, "sendTlv::Error sending val\n");
    }
  }
  else {
    PRINTTRACE(2, SOCKET_LIB_TRACE, "sendTlv::Error sending Type Length\n");
  }
  return ret;
}

//-----------------------------------------------------------------------------
//
//    CONNECTIONS
//
//-----------------------------------------------------------------------------


Connections::Connections(int dbg = 0, char *name = NULL) :
    connectionsHead(NULL), connectionsTail(NULL), numConnections(0), lock(1), count(0), connectionsDebug(dbg)
{
  if (name) 
    strcpy(listName, name);
  else
    strcpy(listName,"");
  if (connectionsDebug) {
    cout << "Connections::Connections" << endl;
    cout << "\tName = " << listName << "; semValue = " << lock.getVal() << "; debug = " << connectionsDebug << endl;
  }
};

void Connections::printMutex(char *pre, char *pre2)
{
  cout << listName << pre << pre2 << " count = " << count << "; semValue = " << lock.getVal() << endl;
};

void Connections::grabMutexLock(char *pre)
{
  if (connectionsDebug)
    printMutex(pre, "::grabMutexLock::");
  lock.wait();
  count++;
};

void Connections::releaseMutexLock(char *pre)
{
  if (connectionsDebug)
    printMutex(pre, "::releaseMutexLock::");
  lock.release();
  count--;
};

void Connections::CleanupConnections()
{
  // Can't grab the mutex here as we already have the mutex locked when the system processed a SIGINT
  // In that case this will prevent the application from exiting.
  //  grabMutexLock("connections::CleanupConnections");

  Connection *p = connectionsHead;
  while(p) {
    Connection *next = p->next;
    delete p;
    p = next;
  }
  connectionsHead = connectionsTail = NULL;
  numConnections = 0;
  //  releaseMutexLock("connections::CleanupConnections");
};

void Connections::findConnectionAndDispatch(fd_set  &readfds, fd_set  &writefds)
{
  int level = 3;
  bool debug = SOCKETS_DEBUG;

  grabMutexLock("connections::findConnectionAndDispatch");
  if (debug) {
    sprintf(traceBuff, "findConnectionAndDispatch:: %s\n", listName);
    printDebug(traceBuff, strlen(traceBuff));
  }
  
  for (Connection *connPtr = connectionsHead; connPtr; connPtr = connPtr->next) {
    int f = connPtr->getFd();
    char stateStr[80];
    char *connectionName = connPtr->getName();
    char *tPtr = traceBuff;

    if (debug) {
      connPtr->stateString(stateStr);
      tPtr += sprintf(tPtr, "\tCONNECTION = %s; fd = %d; state = %s\n",
	      connectionName, f, stateStr);
    }
    if (f == -1) {
      if (debug) {
	tPtr += sprintf(tPtr, "\tignoring\n");
	printDebug(traceBuff, strlen(traceBuff));
      }
      continue;
    }
    if (connPtr->isConnectPending()) {
      bool readfdset = (FD_ISSET(f, &readfds)) ? true : false;
      bool writefdset = (FD_ISSET(f, &writefds)) ? true : false;

      if (debug) {
	tPtr += sprintf(tPtr, "\tConnect Pending\n");
      }
      if (readfdset || writefdset) {
	if (debug) {
	  tPtr += sprintf(tPtr, "\tread fd %s; write fd %s\n",
			  ((readfdset) ? "Set" : "Not Set"),
			  ((writefdset) ? "Set" : "Not Set"));
	}
	int error; socklen_t len = sizeof(error);
	int ret = getsockopt(f, SOL_SOCKET, SO_ERROR, &error, &len);
	const char *errorStr = sys_errlist[errno];
	if (debug) {
	  tPtr += sprintf(tPtr, "\tgetsockopt returned %d; error = %s\n", ret, errorStr);
	}
	if (ret < 0) {
	  if (errno == ECONNREFUSED || errno == ETIMEDOUT) {
	    if (debug) {
	      tPtr += sprintf(tPtr, "\tConnection refused/timed out\n");
	    }
	    connPtr->State = DISCONNECTED;
	    ::close(f);
	    connPtr->setFd(-1);
	  }
	  if (debug) {
	    connPtr->stateString(stateStr);
	    tPtr += sprintf(tPtr, "\tNew State = %s\n", stateStr);
	    printDebug(traceBuff, strlen(traceBuff));
	  }
	  continue;
	}
	else if (ret == 0) {
	  //	  if (errno == EINPROGRESS) {
	  //	    printf("Connection in progress on %d; errno = %d\n", f, errno);
	  //	  }
	  //	  else if (error) {
	  if (error) {
	    ::close(f);
	    connPtr->setFd(-1);
	    connPtr->State = DISCONNECTED;  // Still usable for "client" case
	    connPtr->stateString(stateStr);
	    if (debug) {
	      tPtr += sprintf(tPtr, "\tERROR Connecting on %d; cleaning up\n", f);
	      tPtr += sprintf(tPtr, "\tNew State = %s\n", stateStr);
	      printDebug(traceBuff, strlen(traceBuff));
	    }
	    continue;
	  }
	  else if (errno == EINPROGRESS) {
	    if (debug) {
	      tPtr += sprintf(tPtr, "\tConnect in progress\n");
	    }
	    if (writefdset) {
	      // printf("Write FD set on connection: %d\n", connPtr->getFd());
	      if (debug) {
		tPtr += sprintf(tPtr, "\tConnection Established on fd = %d; connection = %s\n", f, connectionName);
	      }
	      connPtr->connected();
	      connPtr->userConnectionEstablished();  // Call virtual function
	    }
	  }
	}
	if (debug) {
	  connPtr->stateString(stateStr);
	  tPtr += sprintf(tPtr, "\tNew State = %s\n", stateStr);
	  printDebug(traceBuff, strlen(traceBuff));
	}
	continue;
      }
      else {
	PRINTTRACE(level, SOCKET_LIB_TRACE, "\tread/write fds NOT set; - NOTHING TO BE DONE\n");
      }
    }

    if (FD_ISSET(f, &readfds)) {
      if (debug) {
	tPtr += sprintf(tPtr, "\tRead FD set\n");
      }

      if (connPtr->bufferExceeded()) {
 	connPtr->disconnected();
	if (debug) {
	  tPtr += sprintf(tPtr, "\tbuffer exceeded while reading\n");
	  connPtr->stateString(stateStr);
	  tPtr += sprintf(tPtr, "\tNew State = %s\n", stateStr);
	  printDebug(traceBuff, strlen(traceBuff));
	}
 	continue;
      }
      if (!connPtr->read2K()) {
	if (debug) {
	  tPtr += sprintf(tPtr, "\tError reading\n");
	  connPtr->stateString(stateStr);
	  tPtr += sprintf(tPtr, "\tNew State = %s\n", stateStr);
	  printDebug(traceBuff, strlen(traceBuff));
	}
 	continue;
      }
      if (debug) {
	tPtr += sprintf(tPtr, "\tRead data from socket %d; processing...\n", f);
      }
      while (connPtr->getTlv()) { // Got a TLV
	connPtr->processTLV();  // Call virtual function
	if (!connPtr->shrinkReadBuffer()) {
	  if (debug) {
	    tPtr += sprintf(tPtr, "\tError shrinking read buffer - disconnecting;\n");
	    printDebug(traceBuff, strlen(traceBuff));
	  }
	  connPtr->disconnected();
	  break;
	}
      }
    }
    // Get the fd again as getTlv() call could have resulted in disconnect being called 
    f = connPtr->getFd();
    if (f == -1) {
      if (debug) {
	tPtr += sprintf(tPtr, "\tfd = -1; must have disconnected\n");
	connPtr->stateString(stateStr);
	tPtr += sprintf(tPtr, "\tNew State = %s\n", stateStr);
	printDebug(traceBuff, strlen(traceBuff));
      }
      continue;
    }
    if (FD_ISSET(f, &writefds)) {
      //  Used to be like this when client/server side connections were handled in the same place
      //      if ((connPtr->isClient() && connPtr->isConnectPending()) ||
      //	  (connPtr->isServer() && !connPtr->isConnected())) {
      if (!connPtr->isConnected()) {
	if (debug) {
	  tPtr += sprintf(tPtr, "\tConnection Established on fd = %d; connection = %s\n", f, connectionName);
	}
	connPtr->connected();
	connPtr->userConnectionEstablished();  // Call virtual function
      }
      else if (connPtr->State == SOCKET_FULL) {
	// File descriptor is writable
	if (debug) {
	  tPtr += sprintf(tPtr, "\tSocket full flushing...\n");
	}
	connPtr->State = CONNECTED;
	//	connPtr->flush();  
	if (!connPtr->flush()) {  // this could make it go back to SOCKET_FULL
	  if (debug) {
	    sprintf(traceBuff, "\tError - flush failed - disconnecting\n");
	    printDebug(traceBuff, strlen(traceBuff));
	  }
	  connPtr->disconnected();
	}
      }
      else {
	// File descriptor is writable
	if (debug) {
	  tPtr += sprintf(tPtr, "\tSocket OK to Send...\n");
	}
	connPtr->userOkToSend();  // Call virtual function
      }
    }
    PRINTTRACE(level, SOCKET_LIB_TRACE, traceBuff);
  }
  releaseMutexLock("connections::findConnectionAndDispatch");
  if (debug) {
    sprintf(traceBuff, "\tExiting:: %s\n", listName);
    printDebug(traceBuff, strlen(traceBuff));
  }
};

int Connections::setFdSet(fd_set  &fds)
{
  grabMutexLock("connections::setFdSet");
  
  int maxfd = 0;
  for (Connection *p = connectionsHead; p; p = p->next) {
    int f = p->setFdSet(fds);
    if (f != -1) {
      maxfd = (maxfd < f) ? f : maxfd;
    }
  }
  releaseMutexLock("connections::setFdSet");
  return maxfd;
};

int Connections::setReadWriteFdSets(fd_set  &readfds, fd_set  &writefds)
{
  int level = 3;
  bool debug = SOCKETS_DEBUG;

  grabMutexLock("connections::setReadWriteFdSetForServerSockets");
  
  if (debug) {
    sprintf(traceBuff, "setReadWriteFdSets:: %s\n", listName);
    printDebug(traceBuff, strlen(traceBuff));
  }

  int maxfd = 0;
  for (Connection *p = connectionsHead; p; p = p->next) {
    char *connectionName = p->getName();
    int f = p->getFd();
    char stateStr[80];

    if (f == -1)
      continue;
    char *tPtr = traceBuff;
    if (debug) {
      p->stateString(stateStr);
      tPtr += sprintf(tPtr, "\tCONNECTION = %s; fd = %d; state = %s\n",
		      connectionName, f, stateStr);
    }
    if (p->isServer()) {
      FD_SET(f, &readfds);
      if (debug) {
	tPtr += sprintf(tPtr, "\tServer side socket\n");
	tPtr += sprintf(tPtr, "\tSetting read fd %d, in read fdset\n", f);
      }
      if (!p->isConnected() || p->isSocketFull()) {
	// To make connection "CONNECTED"
	FD_SET(f, &writefds);
	if (debug) {
	  tPtr += sprintf(tPtr, "\tNot Connected\n");
	  tPtr += sprintf(tPtr, "\tSetting write fd %d in write fdset\n", f);
	}
      }
      maxfd = (maxfd < f) ? f : maxfd;
    }
    else {
      if (debug) {
	tPtr += sprintf(tPtr, "\tClient side socket\n");
      }
      if (p->isConnectPending() || p->isSocketFull()) {
	if (debug) {
	  tPtr += sprintf(tPtr, "\tConnectPending / SocketFull\n");
	  tPtr += sprintf(tPtr, "\tSetting read/write fd %d in read/write fdset\n", f);
	}
	FD_SET(f, &writefds);
	FD_SET(f, &readfds);
	maxfd = (maxfd < f) ? f : maxfd;
      }
      else if (p->isConnected()) {
	if (debug) {
	  tPtr += sprintf(tPtr, "\tConnected\n");
	  tPtr += sprintf(tPtr, "\tSetting read fd %d in read fdset\n", f);
	}
	FD_SET(f, &readfds);
	maxfd = (maxfd < f) ? f : maxfd;
      }
      else {
	if (debug) {
	  tPtr += sprintf(tPtr, "\tNot Setting fd %d in read or write fdset\n", f);
	}
      }
    }
    PRINTTRACE(level, SOCKET_LIB_TRACE, traceBuff);
  }
  releaseMutexLock("connections::setReadWriteFdSetForServerSockets");
  return maxfd;
};

void Connections::deleteDeadConnections()
{
  grabMutexLock("connections::deleteDeadConnections");
  for (Connection *p = connectionsHead; p;) {
    if (p->isServer() && p->dead()) {
      if (p->checkRefCount() == 0) {
	Connection *temp = p->next;
	sprintf(traceBuff, "Deleting dead connection %s\n", p->getName());
	PRINTTRACE(5, SOCKET_LIB_TRACE, traceBuff);
	delete p;
	// cout << "Deleting Dead connection" << endl;
	p = temp;
	continue;
      }
    }
    p = p->next;
  }
  releaseMutexLock("connections::deleteDeadConnections");
};


void Connections::removeAllChildren()
{
  grabMutexLock("connections::deleteAllChildren");
  connectionsHead = NULL;
  connectionsTail = NULL;
  releaseMutexLock("connections::deleteAllChildren");
};


void Connections::add(Connection *conn)
{
  if (!conn) return;
  grabMutexLock("connections::add");
  
  sprintf(traceBuff, "Adding connection %s to the connections list %s\n", conn->getName(), listName);
  PRINTTRACE(5, SOCKET_LIB_TRACE, traceBuff);

  conn->next = NULL; conn->prev = connectionsTail; 
  if (connectionsTail) {connectionsTail->next = conn; connectionsTail = conn;}
  if (!connectionsHead) {connectionsHead = connectionsTail = conn;}
  numConnections++;
  releaseMutexLock("connections::add");
};

void Connections::remove(Connection *conn)
{
  if (!conn) return;
  //  Cant grab mutex deleteDeadConnections->delete conn->remove
  //  grabMutexLock("connections::remove");
  sprintf(traceBuff, "Removing connection %s from the connections list %s\n", conn->getName(), listName);
  PRINTTRACE(5, SOCKET_LIB_TRACE, traceBuff);
  
  if (conn->prev) {conn->prev->next = conn->next;}
  if (conn->next) {conn->next->prev = conn->prev;}
  if (connectionsHead == conn) {connectionsHead = conn->next;}
  if (connectionsTail == conn) {connectionsTail = conn->prev;}
  numConnections--;
  //  releaseMutexLock("connections::remove");
};


//-----------------------------------------------------------------------------
//
//    CONNECTION
//
//-----------------------------------------------------------------------------

Connection::Connection(Connections *list, int clientOrServer, char *name) : 
  SocketInterface(clientOrServer), connectionsList(list), lock(0)
{
  next = prev = NULL;
  if (name == NULL) 
    strcpy(connName, "default");
  else
    strcpy(connName, name);

  if (connectionsList)
    connectionsList->add(this);
};

Connection::Connection(Connections *list, int clientOrServer, int fd, char *name) :
  SocketInterface(clientOrServer, fd), connectionsList(list), lock(0)
{
  next = prev = NULL;
  if (name == NULL) 
    strcpy(connName, "default");
  else
    strcpy(connName, name);

  if (connectionsList)
    connectionsList->add(this);
};

Connection::~Connection()
{
  socketCleanup(); //
  if (connectionsList) connectionsList->remove(this);
};

void Connection::disconnected()
{
  // remove() called from the destructor
  userDisconnectProcessing();
  socketCleanup(); //
};

int Connection::setFdSet(fd_set  &fds)
{
  int f = getFd();
  if (f == -1) return 0;
  FD_SET(f, &fds);
  return f;
};

int Connection::Connect(struct sockaddr *serv_addr)
{
  int n;
  int fd = getFd();
  int ret = 0;

  if (fd == -1) {
    cout << "Connection::Connect:: Bad file descriptor" << endl;
  }
  else {
    //    cout << "Connection::Connect:: connecting ...; fd = " << fd << endl;
    int flags = fcntl(fd, F_GETFL, 0);
    fcntl(fd, F_SETFL, flags | O_NONBLOCK);

    if ((n = connect(fd, (struct sockaddr *)serv_addr, sizeof(*serv_addr))) < 0) {
      int locErr = errno;
      if (locErr == EINPROGRESS) {
	// perror("Connection::");
	connectPending();
	time(&lastConnectTime);
      }
      else if (locErr == EISCONN) {
	//      printf("Errno = %d Str = %s\n", locErr, sys_errlist[locErr]);
	// perror("Connection::");
	connected();
	ret = 1;  // success
      }
      else {
	//	perror("Connect Error - retry later"); //-- Commented out because this happens too often 
      }
    }
    else {
      connected();
      ret = 1;  // success
    }
  }
  //  cout << "socketInterface::Connect returning " << ((ret == 1) ? "success" : "failure") << endl;
  return ret; 
}


